(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('@yellicode/model'), require('@yellicode/templating')) :
	typeof define === 'function' && define.amd ? define(['exports', '@yellicode/model', '@yellicode/templating'], factory) :
	(factory((global['@yellicode/typescript'] = {}),global.model,global.templating));
}(this, (function (exports,model,templating) { 'use strict';

/**
 * Defines options for dealing with elements that are 'optional', meaning that they have a lower bound of 0.
 */

(function (OptionalityModifier) {
    /**
     * Don't write any modifier.
     */
    OptionalityModifier[OptionalityModifier["Ignore"] = 0] = "Ignore";
    /**
     * Write a 'null' keyword, for example 'myProperty: string | null;';
     */
    OptionalityModifier[OptionalityModifier["NullKeyword"] = 1] = "NullKeyword";
    /**
     * Write a question token, for example 'myProperty?: string;';
     */
    OptionalityModifier[OptionalityModifier["QuestionToken"] = 2] = "QuestionToken";
})(exports.OptionalityModifier || (exports.OptionalityModifier = {}));

(function (EnumFeatures) {
    EnumFeatures[EnumFeatures["None"] = 0] = "None";
    EnumFeatures[EnumFeatures["JsDocDescription"] = 1] = "JsDocDescription";
    EnumFeatures[EnumFeatures["Initializers"] = 2] = "Initializers";
    EnumFeatures[EnumFeatures["All"] = 3] = "All";
    EnumFeatures[EnumFeatures["AllExceptJsDoc"] = 2] = "AllExceptJsDoc";
})(exports.EnumFeatures || (exports.EnumFeatures = {}));

(function (ClassFeatures) {
    ClassFeatures[ClassFeatures["None"] = 0] = "None";
    ClassFeatures[ClassFeatures["JsDocDescription"] = 1] = "JsDocDescription";
    ClassFeatures[ClassFeatures["Generalizations"] = 2] = "Generalizations";
    ClassFeatures[ClassFeatures["InterfaceRealizations"] = 4] = "InterfaceRealizations";
    ClassFeatures[ClassFeatures["All"] = 7] = "All";
    ClassFeatures[ClassFeatures["AllExceptJsDoc"] = 6] = "AllExceptJsDoc";
})(exports.ClassFeatures || (exports.ClassFeatures = {}));

(function (InterfaceFeatures) {
    InterfaceFeatures[InterfaceFeatures["None"] = 0] = "None";
    InterfaceFeatures[InterfaceFeatures["JsDocDescription"] = 1] = "JsDocDescription";
    InterfaceFeatures[InterfaceFeatures["Generalizations"] = 2] = "Generalizations";
    InterfaceFeatures[InterfaceFeatures["All"] = 3] = "All";
    InterfaceFeatures[InterfaceFeatures["AllExceptJsDoc"] = 2] = "AllExceptJsDoc";
})(exports.InterfaceFeatures || (exports.InterfaceFeatures = {}));

(function (PropertyFeatures) {
    PropertyFeatures[PropertyFeatures["None"] = 0] = "None";
    PropertyFeatures[PropertyFeatures["JsDocDescription"] = 1] = "JsDocDescription";
    PropertyFeatures[PropertyFeatures["AccessModifier"] = 2] = "AccessModifier";
    PropertyFeatures[PropertyFeatures["ReadonlyModifier"] = 4] = "ReadonlyModifier";
    PropertyFeatures[PropertyFeatures["OptionalModifier"] = 8] = "OptionalModifier";
    // Initializer = 1 << 4,
    PropertyFeatures[PropertyFeatures["All"] = 15] = "All"; /*| Initializer */
    PropertyFeatures[PropertyFeatures["AllExceptJsDoc"] = 14] = "AllExceptJsDoc"; /* | Initializer*/
})(exports.PropertyFeatures || (exports.PropertyFeatures = {}));

(function (FunctionFeatures) {
    FunctionFeatures[FunctionFeatures["None"] = 0] = "None";
    FunctionFeatures[FunctionFeatures["Comments"] = 1] = "Comments";
    FunctionFeatures[FunctionFeatures["JsDocDescription"] = 2] = "JsDocDescription";
    FunctionFeatures[FunctionFeatures["JsDocParameters"] = 4] = "JsDocParameters";
    FunctionFeatures[FunctionFeatures["AccessModifier"] = 8] = "AccessModifier";
    FunctionFeatures[FunctionFeatures["All"] = 15] = "All";
    FunctionFeatures[FunctionFeatures["AllExceptJsDoc"] = 8] = "AllExceptJsDoc";
})(exports.FunctionFeatures || (exports.FunctionFeatures = {}));

class TypeScriptTypeNameProvider extends templating.DefaultTypeNameProvider {
    getDataTypeName(type) {
        if (model.isPrimitiveBoolean(type))
            return "boolean";
        if (model.isPrimitiveInteger(type) || model.isPrimitiveReal(type))
            return "number";
        if (model.isPrimitiveString(type))
            return "string";
        if (model.isPrimitiveObject(type))
            return "any";
        return super.getDataTypeName(type);
    }
}

/**
 * Provides code writing functionality specific for TypeScript.
 */
class TypeScriptWriter extends templating.CodeWriter {
    constructor(writer, options) {
        super(writer);
        if (!options)
            options = {};
        this.typeNameProvider = options.typeNameProvider || new TypeScriptTypeNameProvider();
        this.maxCommentWidth = options.maxCommentWidth || 100;
    }
    /**
    * Writes an indented block of code, wrapped in opening and closing brackets.
    * @param contents A callback function that writes the contents.
    */
    writeCodeBlock(contents) {
        this.writeLine('{');
        this.increaseIndent();
        if (contents)
            contents(this);
        this.decreaseIndent();
        this.writeLine('}');
    }
    ;
    /**
     * Writes a block of code, wrapped in a class declaration and opening and closing brackets.
     * This function does not write class members.
     * @param cls The class.
     * @param contents A callback function that writes the class contents.
     * @param options An optional ClassOptions object.
     */
    writeClassBlock(cls, contents, options) {
        if (!cls)
            return;
        if (!options)
            options = {};
        const features = (options.features === undefined) ? exports.ClassFeatures.All : options.features;
        if (features & exports.ClassFeatures.JsDocDescription) {
            this.writeJsDocDescription(cls.ownedComments);
        }
        this.writeIndent();
        if (options.prefix) {
            this.write(`${options.prefix.trim()} `);
        }
        if (cls.isAbstract) {
            this.write('abstract ');
        }
        this.write(`class ${cls.name}`);
        if (features & exports.ClassFeatures.Generalizations) {
            this.writeGeneralizations(cls.generalizations, options.inherits);
        }
        if (features & exports.ClassFeatures.InterfaceRealizations) {
            this.writeInterfaceRealizations(cls.interfaceRealizations, options.implements);
        }
        this.writeEndOfLine();
        this.writeCodeBlock((writer) => { contents(writer, cls); });
    }
    /**
      * Writes a block of code, wrapped in an interface declaration and opening and closing brackets.
      * This function does not write interface members.
      * @param iface The interface.
      * @param contents A callback function that writes the interface contents.
      * @param options An optional InterfaceOptions object.
      */
    writeInterfaceBlock(iface, contents, options) {
        if (!iface)
            return;
        if (!options)
            options = {};
        const features = (options.features === undefined) ? exports.InterfaceFeatures.All : options.features;
        if (features & exports.InterfaceFeatures.JsDocDescription) {
            this.writeJsDocDescription(iface.ownedComments);
        }
        this.writeIndent();
        if (options.prefix) {
            this.write(`${options.prefix.trim()} `);
        }
        this.write(`interface ${iface.name}`);
        if (features & exports.InterfaceFeatures.Generalizations) {
            this.writeGeneralizations(iface.generalizations, options.inherits);
        }
        this.writeEndOfLine();
        this.writeCodeBlock((writer) => { contents(writer, iface); });
    }
    /**
     * Writes a class or interface property.
     * @param property The property to write.
     * @param options An optional PropertyOptions object.
     */
    writeProperty(property, options) {
        if (!property)
            return;
        if (!options)
            options = {};
        const features = (options.features === undefined) ? exports.PropertyFeatures.All : options.features;
        const optionalityModifier = (options.optionality === undefined) ? exports.OptionalityModifier.QuestionToken : options.optionality;
        const makeOptional = property.isOptional() && (features & exports.PropertyFeatures.OptionalModifier);
        if (features & exports.PropertyFeatures.JsDocDescription) {
            this.writeJsDocDescription(property.ownedComments);
        }
        // Start a new, indented line        
        this.writeIndent();
        // Access modifier (+ white space)
        if ((features & exports.PropertyFeatures.AccessModifier) && !model.isInterface(property.owner)) {
            this.writeAccessModifier(property.visibility);
        }
        // Readonly modifier
        if ((features & exports.PropertyFeatures.ReadonlyModifier) && property.isReadOnly || property.isDerived) {
            this.write('readonly ');
        }
        // The name
        this.write(property.name);
        if (makeOptional && (optionalityModifier & exports.OptionalityModifier.QuestionToken)) {
            this.write('?');
        }
        // The type        
        this.write(`: ${this.getTypeName(property.type) || 'any'}`);
        if (property.isMultivalued()) {
            this.write('[]');
        }
        if (makeOptional && (optionalityModifier & exports.OptionalityModifier.NullKeyword)) {
            this.write(' | null');
        }
        // TODO: Initializer if not an interface and the model has a default value (but should have a some optional value conversion hook, primitives only)
        // if ((features & opts.PropertyFeatures.Initializer) && !model.isInterface(property.owner)) {
        // }
        this.writeEndOfLine(';');
    }
    /**
     * Writes a full enumeration, including members.
     * @param element The enumeration.
     * @param options An optional EnumerationOptions object.
     */
    writeEnumeration(enumeration, options) {
        if (!enumeration)
            return;
        if (!options)
            options = {};
        const features = (options.features === undefined) ? exports.EnumFeatures.All : options.features;
        if (features & exports.EnumFeatures.JsDocDescription) {
            this.writeJsDocDescription(enumeration.ownedComments);
        }
        this.writeIndent();
        if (options.prefix) {
            this.write(`${options.prefix.trim()} `);
        }
        this.write(`enum ${enumeration.name}`);
        this.writeEndOfLine();
        this.writeCodeBlock(() => {
            let literals = enumeration.ownedLiterals;
            if (!literals)
                return;
            for (let i = 0, len = literals.length; i < len; i++) {
                const literal = literals[i];
                if (features & exports.EnumFeatures.JsDocDescription) {
                    this.writeJsDocDescription(literal.ownedComments);
                }
                this.writeIndent();
                this.write(literal.name);
                if ((features & exports.EnumFeatures.Initializers) && literal.specification) {
                    let initialValue = literal.specification.getStringValue();
                    if (model.isLiteralString(literal.specification)) {
                        initialValue = `'${initialValue}'`; // a string enum: wrap in quotes
                    }
                    this.write(` = ${initialValue}`);
                }
                if (i < len - 1) {
                    this.write(',');
                }
                this.writeEndOfLine();
            }
        });
    }
    /**
     * Writes a string literal type from a specified enumeration. Example: 'type Easing = 'ease-in' | 'ease-out' | 'ease-in-out';'
     * @param enumeration The enumeration.
     * @param prefix An optional prefix, such as 'export'.
     */
    writeStringLiteralType(enumeration, prefix) {
        this.writeJsDocDescription(enumeration.ownedComments);
        this.writeIndent();
        if (prefix && prefix.length > 0) {
            this.write(`${prefix.trim()} `);
        }
        this.write(`type ${enumeration.name} = `);
        this.joinWrite(enumeration.ownedLiterals, ' | ', lit => `'${lit.name}'`);
        this.writeEndOfLine(';');
    }
    /**
     * Writes a function declaration without a body.
     * @param operation The operation.
     * @param options An optional FunctionOptions object.
     */
    writeFunctionDeclaration(operation, options) {
        if (!operation)
            return;
        this.writeFunctionStart(operation, options);
        this.writeEndOfLine(';');
    }
    /**
    * Writes a block of code, wrapped in an function declaration and opening and closing brackets.      *
    * @param operation The operation.
    * @param contents A callback that writes the operation contents.
    * @param options An optional FunctionOptions object.
    */
    writeFunctionBlock(operation, contents, options) {
        if (!operation)
            return;
        this.writeFunctionStart(operation, options);
        if (operation.isAbstract) {
            this.writeEndOfLine(';');
            return;
        }
        this.writeEndOfLine();
        this.writeCodeBlock((writer) => { contents(writer, operation); });
    }
    writeFunctionStart(operation, options) {
        if (!operation)
            return;
        if (!options)
            options = {};
        const features = (options.features === undefined) ? exports.FunctionFeatures.All : options.features;
        // jsDoc tags 
        var jsDocLines = [];
        if (features & exports.FunctionFeatures.JsDocDescription) {
            this.pushJsDocLinesFromComments(operation.ownedComments, jsDocLines);
        }
        if (features & exports.FunctionFeatures.JsDocParameters) {
            this.pushJsDocLinesForParameters(operation.ownedParameters, jsDocLines);
        }
        this.writeJsDocLines(jsDocLines);
        // Start a new, indented line
        this.writeIndent();
        if (!model.isInterface(operation.owner)) {
            this.writeAccessModifier(operation.visibility);
        }
        if (operation.isStatic) {
            this.write('static ');
        }
        else if (operation.isAbstract) {
            this.write('abstract ');
        }
        this.write(operation.name);
        // If needed, make the function optional using the '?'   
        const returnParameter = operation.getReturnParameter();
        const optionalityModifier = (options.returnOptionality === undefined) ? exports.OptionalityModifier.NullKeyword : options.returnOptionality;
        const makeOptional = (returnParameter && returnParameter.isOptional());
        if (makeOptional && (optionalityModifier & exports.OptionalityModifier.QuestionToken)) {
            this.write('?');
        }
        this.write('(');
        this.writeInOutParameters(operation.ownedParameters, options.parameterOptionality);
        this.write('): ');
        // Write the return type        
        const returnType = this.getParameterTypeName(returnParameter) || 'void';
        this.write(returnType);
        if (makeOptional && (optionalityModifier & exports.OptionalityModifier.NullKeyword)) {
            this.write(' | null');
        }
    }
    writeInOutParameters(elements, optionalityModifier) {
        if (!elements)
            return;
        if (optionalityModifier === undefined)
            optionalityModifier = exports.OptionalityModifier.NullKeyword;
        let i = 0;
        elements.forEach((p) => {
            if (p.direction === model.ParameterDirectionKind.return)
                return;
            const makeOptional = p.isOptional();
            if (i > 0) {
                this.write(', ');
            }
            this.write(p.name);
            if (makeOptional && (optionalityModifier & exports.OptionalityModifier.QuestionToken)) {
                this.write('?');
            }
            const typeName = this.getParameterTypeName(p) || 'any';
            this.write(`: ${typeName}`);
            if (makeOptional && (optionalityModifier & exports.OptionalityModifier.NullKeyword)) {
                this.write(' | null');
            }
            i++;
        });
    }
    getParameterTypeName(parameter) {
        if (!parameter)
            return null;
        var typeName = this.getTypeName(parameter.type) || 'any';
        if (parameter.isMultivalued()) {
            typeName = `${typeName}[]`;
        }
        return typeName;
    }
    writeAccessModifier(visibilityKind) {
        const visibilityString = TypeScriptWriter.getAccessModifierString(visibilityKind);
        if (!visibilityString)
            return;
        this.write(visibilityString);
        this.writeWhiteSpace();
    }
    writeGeneralizations(generalizations, additional) {
        const allNames = [];
        if (generalizations) {
            allNames.push(...generalizations.map(g => this.getTypeName(g.general)));
        }
        if (additional) {
            allNames.push(...additional);
        }
        if (allNames.length === 0)
            return;
        this.write(' extends ');
        this.joinWrite(allNames, ', ', name => name);
    }
    writeInterfaceRealizations(realizations, additional) {
        const allNames = [];
        if (realizations) {
            allNames.push(...realizations.map(ir => this.getTypeName(ir.contract)));
        }
        if (additional) {
            allNames.push(...additional);
        }
        if (allNames.length === 0)
            return;
        this.write(' implements ');
        this.joinWrite(allNames, ', ', name => name);
    }
    pushJsDocLinesFromComments(comments, lines) {
        if (!comments)
            return;
        let i = 0;
        comments.forEach((p) => {
            if (!p.body) {
                return;
            }
            if (i > 0) {
                lines.push(''); // add a blank line
            }
            lines.push(p.body);
            i++;
        });
    }
    pushJsDocLinesForParameters(parameters, lines) {
        if (!parameters)
            return;
        parameters.forEach((p) => {
            lines.push(this.getJsDocLineForParameter(p));
        });
    }
    getJsDocLineForParameter(parameter) {
        const isReturn = parameter.direction === model.ParameterDirectionKind.return;
        const tag = isReturn ? 'returns' : 'param';
        const commentBodies = parameter.ownedComments.map(c => c.body);
        let typeName = this.getTypeName(parameter.type) || 'any';
        if (parameter.isMultivalued()) {
            typeName = `${typeName}[]`;
        }
        let line = `@${tag} {${typeName}}`;
        if (!isReturn) {
            line = `${line} ${parameter.name}`;
        }
        if (commentBodies.length > 0) {
            line = `${line} ${commentBodies.join(' ')}`;
        }
        return line;
    }
    writeJsDocDescription(data) {
        const lines = [];
        if (typeof data == 'string') {
            lines.push(data);
        }
        else {
            this.pushJsDocLinesFromComments(data, lines);
        }
        this.writeJsDocLines(lines);
    }
    writeJsDocParagraph(text) {
        this.writeJsDocLines([text]);
    }
    writeJsDocLines(lines) {
        if (lines.length === 0)
            return;
        this.writeLine('/**');
        // Split here
        lines.forEach(line => {
            const lineLength = line ? line.length : 0;
            if (this.maxCommentWidth > 0 && lineLength > this.maxCommentWidth) {
                // See if we can split the line
                var split = templating.CodeWriterUtility.wordWrap(line, this.maxCommentWidth);
                split.forEach(s => {
                    this.writeLine(`* ${s}`);
                });
            }
            else
                this.writeLine(`* ${line}`);
        });
        this.writeLine('*/');
    }
    static getAccessModifierString(visibility) {
        switch (visibility) {
            case model.VisibilityKind.public:
                return 'public';
            case model.VisibilityKind.private:
                return 'private';
            case model.VisibilityKind.protected:
                return 'protected';
            default:
                return null;
        }
    }
    getTypeName(type) {
        if (!type)
            return null;
        if (!this.typeNameProvider)
            return type.name;
        return this.typeNameProvider.getTypeName(type);
    }
    joinWrite(collection, separator, getStringFunc) {
        let isFirst = true;
        collection.forEach(c => {
            const value = getStringFunc(c);
            if (!value)
                return;
            if (isFirst) {
                isFirst = false;
            }
            else
                this.write(separator);
            this.write(value);
        });
    }
}

exports.TypeScriptTypeNameProvider = TypeScriptTypeNameProvider;
exports.TypeScriptWriter = TypeScriptWriter;

Object.defineProperty(exports, '__esModule', { value: true });

})));
